/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a generic SCSI interface to Mustek scanners.  */

#include <alloca.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <scsi/sg.h>

#include <sys/ioctl.h>
#include <sys/param.h>

#include <genscsi.h>
#include <sanei.h>

SANE_Status
genscsi_open (const char * dev, int * fdp)
{
  int fd, timeout;

  fd = open(dev, O_RDWR | O_EXCL);
  if (fd < 0)
    return SANE_STATUS_INVAL;

  /* Set large timeout since some scanners are slow but do not
     disconnect... ;-( */
  timeout = 10*60*HZ;		/* how about 10 minutes? ;-) */
  ioctl(fd, SG_SET_TIMEOUT, &timeout);

  if (fdp)
    *fdp = fd;
  return SANE_STATUS_GOOD;
}


SANE_Status
genscsi_cmd (int fd, const void * src, size_t src_size,
	  void * dst, size_t * dst_size)
{
  /* The Linux generic scsi interface is braindead as it forces us to
     copy things around like mad.  Fortunately, most command requests
     are small.  */
  struct
  {
    struct sg_header	hdr;
    u_int8_t		data[SG_BIG_BUFF];
  } * req;
  ssize_t nread;

  req = alloca(sizeof(*req));	/* works around gdb bug... ;-( */

  memset(&req->hdr, 0, sizeof(req->hdr));
  req->hdr.pack_len  = src_size + sizeof(req->hdr);
  req->hdr.reply_len = (dst_size ? *dst_size : 0) + sizeof(req->hdr);
  memcpy(&req->data, src, src_size);

  if (write(fd, req, req->hdr.pack_len) != req->hdr.pack_len)
    {
      DBG(1, "genscsi_cmd: bad write\n");
      return SANE_STATUS_IO_ERROR;
    }

  nread = read(fd, req, req->hdr.reply_len);
  if (nread < 0)
    {
      DBG(1, "genscsi_cmd: read returned %ld (errno=%d)\n",
	  (long) nread, errno);
      return SANE_STATUS_IO_ERROR;
    }

  nread -= sizeof(req->hdr);

  if (req->hdr.result != 0)
    {
      DBG(1, "genscsi_cmd: SCSI completed with result=%d\n", req->hdr.result);
      return SANE_STATUS_IO_ERROR;
    }

  if (dst)
    memcpy(dst, req->data, nread);

  if (dst_size)
    *dst_size = nread;

  return SANE_STATUS_GOOD;
}

void
genscsi_close (int fd)
{
  close(fd);
}
