/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a SANE backend for Mustek flatbed scanners.  */

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <sane.h>
#include <sanei.h>
#include <saneopts.h>
#include <genscsi.h>
#include <mustek.h>

#define PATH_MUSTEK_CONFIG	PATH_SANE_CONFIG_DIR "/mustek.conf"

#define MM_PER_INCH	25.4

#define NELEMS(a)	((int) sizeof(a)/sizeof(a[0]))

static int num_devices;
static Mustek_Device * first_dev;

static const SANE_String_Const mode_list[] =
{
  "Lineart", "Halftone", "Gray",
  "Color Lineart", "Color Halftone", "Color",
  0
};

static const SANE_String_Const speed_list[] =
{
  "Slowest", "Slower", "Normal", "Faster", "Fastest",
  0
};

static const SANE_String_Const source_list[] =
{
  "Flatbed", "Automatic Document Feeder", "Transparency Adapter",
  0
};

static const SANE_Int grain_list[] =
{
  6,				/* # of elements */
  2, 3, 4, 5, 6, 8
};

static const SANE_Range u8_range =
{
    0,				/* minimum */
  255,				/* maximum */
    0				/* quantization */
};

static const SANE_Int pattern_dimension_list[] =
{
  8,				/* # of elements */
  0, 2, 3, 4, 5, 6, 7, 8
};

static const SANE_Range percentage_range =
{
  -100 << SANE_FIXED_SCALE_SHIFT,	/* minimum */
   100 << SANE_FIXED_SCALE_SHIFT,	/* maximum */
     1 << SANE_FIXED_SCALE_SHIFT	/* quantization */
};

/* Color band codes: */
#define MUSTEK_CODE_GRAY		0
#define MUSTEK_CODE_RED			1
#define MUSTEK_CODE_GREEN		2
#define MUSTEK_CODE_BLUE		3

/* SCSI commands that the Mustek scanners understand: */
#define MUSTEK_SCSI_TEST_UNIT_READY	0x00
#define MUSTEK_SCSI_REQUEST_SENSE	0x03
#define MUSTEK_SCSI_AREA_AND_WINDOWS	0x04
#define MUSTEK_SCSI_READ_SCANNED_DATA	0x08
#define MUSTEK_SCSI_GET_IMAGE_STATUS	0x0f
#define MUSTEK_SCSI_ADF_AND_BACKTRACK	0x10
#define MUSTEK_SCSI_CCD_DISTANCE	0x11
#define MUSTEK_SCSI_INQUIRY		0x12
#define MUSTEK_SCSI_MODE_SELECT		0x15
#define MUSTEK_SCSI_START_STOP		0x1b
#define MUSTEK_SCSI_LOOKUP_TABLE	0x55

#define INQ_LEN	0x60
static const u_int8_t inquiry[] =
{
  MUSTEK_SCSI_INQUIRY, 0x00, 0x00, 0x00, INQ_LEN, 0x00
};

static const u_int8_t test_unit_ready[] =
{
  MUSTEK_SCSI_TEST_UNIT_READY, 0x00, 0x00, 0x00, 0x00, 0x00
};

static const u_int8_t sense[] =
{
  MUSTEK_SCSI_REQUEST_SENSE, 0x00, 0x00, 0x00, 0x04, 0x00
};

static const u_int8_t stop[] =
{
  MUSTEK_SCSI_START_STOP, 0x00, 0x00, 0x00, 0x00, 0x00
};

static const u_int8_t distance[] =
{
  MUSTEK_SCSI_CCD_DISTANCE, 0x00, 0x00, 0x00, 0x05, 0x00
};

static const u_int8_t get_status[] =
{
  MUSTEK_SCSI_GET_IMAGE_STATUS, 0x00, 0x00, 0x00, 0x06, 0x00
};

#define STORE16(p,v)				\
do {						\
    int value = (v);				\
						\
    *(cp)++ = value & 0xff;			\
    *(cp)++ = (value >> 8) & 0xff;		\
} while (0)


#if __GNU_LIBRARY__ - 0 < 6

char *
strndup (const char * s, size_t n)
{
    char * clone;

    clone = malloc(n + 1);
    strncpy(clone, s, n);
    clone[n] = '\0';
    return clone;
}

#endif /* __GNU_LIBRARY__ */


SANE_Status
attach (const char * devname, Mustek_Device ** devp)
{
  u_int8_t result[INQ_LEN];
  const char *model_name = result + 44;
  int fd, mustek_scanner;
  Mustek_Device * dev;
  SANE_Status status;
  size_t size;

  for (dev = first_dev; dev; dev = dev->next)
    if (strcmp (dev->sane.name, devname) == 0)
      {
	if (devp)
	  *devp = dev;
	return SANE_STATUS_GOOD;
      }

  status = genscsi_open (devname, &fd);
  if (status != SANE_STATUS_GOOD)
    return SANE_STATUS_INVAL;

  size = sizeof (result);
  status = genscsi_cmd (fd, inquiry, sizeof (inquiry), result, &size);
  if (status != SANE_STATUS_GOOD || size != INQ_LEN)
    {
      genscsi_close (fd);
      return status;
    }

  status = genscsi_cmd (fd, test_unit_ready, sizeof (test_unit_ready), 0, 0);
  if (status != SANE_STATUS_GOOD)
    {
      genscsi_close (fd);
      return status;
    }

  genscsi_close (fd);

  /* first check for new firmware format: */
  mustek_scanner = (strncmp (result + 8, "SCANNER", 7) == 0
		    && strncmp (result + 36, "MUSTEK", 6) == 0);

  if (!mustek_scanner)
    {
      /* check for old format: */
      mustek_scanner |= (strncmp (result + 8, "MUSTEK", 6) == 0);
      model_name = result + 16;
    }

  if (!mustek_scanner)
    return SANE_STATUS_INVAL;

  dev = malloc (sizeof (*dev));
  if (!dev)
    return SANE_STATUS_NO_MEM;

  memset(dev, 0, sizeof (*dev));

  dev->sane.name   = strdup(devname);
  dev->sane.vendor = "Mustek";
  dev->sane.model  = strndup(model_name, 11);
  dev->sane.type   = "flatbed scanner";

  dev->x_range.min = 0;
  dev->y_range.min = 0;
  dev->x_range.quant = 0;
  dev->y_range.quant = 0;
  dev->dpi_range.min = SANE_FIX (1);
  dev->dpi_range.quant = SANE_FIX (1);
  if (strncmp (model_name, "MFS-12000CX", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (1200);
    }
  else if (strncmp (model_name, "MFS-06000CX", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (600);
    }
  else if (strncmp (model_name, "MFS-12000SP", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (1200);
    }
  else if (strncmp (model_name, "MFS-08000SP", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.50 * MM_PER_INCH);
      dev->y_range.max = SANE_FIX (13.85 * MM_PER_INCH);
      dev->dpi_range.max = SANE_FIX (800);
    }
  else if (strncmp (model_name, "MFS-06000SP", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (600);
    }
  else if (strncmp (model_name, "MFS-08000CZ", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (800);
    }
  else if (strncmp (model_name, "MFS-06000CZ", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (600);
    }
  else if (strncmp (model_name, "MFC-08000CZ", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (8.5 * MM_PER_INCH);  /* is this correct? */
      dev->y_range.max = SANE_FIX (14.0 * MM_PER_INCH); /* is this correct? */
      dev->dpi_range.max = SANE_FIX (800);
    }
  else if (strncmp (model_name, "MFC-06000CZ", 11) == 0)
    {
      dev->x_range.max = SANE_FIX (220.0);	/* measured */
      dev->y_range.max = SANE_FIX (300.0);	/* measured */
      dev->dpi_range.max = SANE_FIX (600);
    }
  else
    {
      free (dev);
      return SANE_STATUS_INVAL;
    }

  if (result[57] & (1 << 6))
    dev->flags |= MUSTEK_FLAG_SINGLE_PASS;
  else
    {
      /* three-pass scanners quantize to 1% of the maximum resolution/2: */
      dev->dpi_range.quant = dev->dpi_range.max / 2 / 100;
      dev->dpi_range.min = dev->dpi_range.quant;
    }
  if (result[63] & (1 << 2))
    dev->flags |= MUSTEK_FLAG_ADF;
  if (result[63] & (1 << 3))
    dev->flags |= MUSTEK_FLAG_ADF_READY;
  if (result[63] & (1 << 6))
    dev->flags |= MUSTEK_FLAG_TA;

  ++num_devices;
  dev->next = first_dev;
  first_dev = dev;

  if (devp)
    *devp = dev;
  return SANE_STATUS_GOOD;
}

static SANE_Status
constrain_value (Mustek_Scanner * s, SANE_Int option, void *value,
		 SANE_Int * info)
{
  SANE_Fixed w, dpi;

  if (option == OPT_RESOLUTION && !(s->hw->flags & MUSTEK_FLAG_SINGLE_PASS))
    {
      /* The three pass scanners use a 1% of the maximum resolution
         increment for resolutions less than or equal to half of the
         maximum resolution and a 10% of the maximum resolution
         increment for larger resolutions.  We can't represent this
         easily in SANE, so the constraint is simply for 1dpi and then
         we round to the 10% increments if necessary.  */
      SANE_Fixed max_dpi, quant, half_res;

      w = *(SANE_Word *) value;
      max_dpi = s->hw->dpi_range.max;
      half_res = max_dpi / 2;

      if (w > half_res)
	{
	  /* round to 10% step: */
	  quant = (half_res / 10);
	  dpi = (w + quant / 2) / quant;
	  dpi *= quant;
	  if (dpi != w)
	    {
	      *(SANE_Word *) value = dpi;
	      if (info)
		*info |= SANE_INFO_INEXACT;
	    }
	}
    }
  return sanei_constrain_value (s->opt + option, value, info);
}

/* Quantize s->req.resolution and return the resolution code for the
   quantized resolution.  Quantization depends on scanner type (single
   pass vs. three-pass).  */
static int
encode_resolution (Mustek_Scanner * s)
{
  SANE_Fixed max_dpi, dpi;
  int code;

  dpi = s->val[OPT_RESOLUTION].w;

  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    code = dpi >> SANE_FIXED_SCALE_SHIFT;
  else
    {
      SANE_Fixed quant, half_res;

      max_dpi = s->hw->dpi_range.max;
      half_res = max_dpi / 2;

      if (dpi <= half_res)
	{
	  quant = half_res / 100;
	  code = (dpi + quant / 2) / quant;
	  if (code < 1)
	    code = 1;
	  dpi = code * quant;
	}
      else
	{
	  /* quantize to 10% step: */
	  quant = (half_res / 10);
	  code = (dpi + quant / 2) / quant;
	  dpi = code * quant;

	  /* and now encode: */
	  code = ((code - 5) * 30) % 100;
	  if (code == 0)
	    code = 0x64;
	  code |= 0x100;	/* indicate 10% quantization */
	}
    }
  return code;
}

static int
encode_percentage (Mustek_Scanner * s, double value, double quant)
{
  int max, code, sign = 0;

  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    {
      if (value < 0.0)
	{
	  value = -value;
	  sign = 0x80;
	}
      code = (int) (value / 100.0 * 127 + 0.5);
      code |= sign;
      max = 0xff;
    }
  else
    {
      code = (int) ((value / 100.0 / quant) + 0.5) + 0x0c;
      max = 0x18;
    }
  if (code > max)
    code = max;
  if (code < 0)
    code = 0x00;
  return code;
}

static SANE_Status
request_sense (Mustek_Scanner * s)
{
  u_int8_t result[4];
  SANE_Status status;
  size_t len;

  len = sizeof (result);
  status = genscsi_cmd (s->fd, sense, sizeof (sense), result, &len);
  if (status != SANE_STATUS_GOOD)
    return status;

  switch (result[0])
    {
    case 0x00:
      break;

    case 0x82:
      if ((s->hw->flags & MUSTEK_FLAG_ADF) && result[1] & 0x80)
	return SANE_STATUS_JAMMED;
      break;

    case 0x83:
      if ((s->hw->flags & MUSTEK_FLAG_ADF) && result[2] & 0x02)
	return SANE_STATUS_NO_DOCS;
      break;

    case 0x84:
      if ((s->hw->flags & MUSTEK_FLAG_TA) && result[1] & 0x10)
	return SANE_STATUS_COVER_OPEN;
      break;

    default:
      return SANE_STATUS_IO_ERROR;
    }
  return SANE_STATUS_GOOD;
}

static SANE_Status
scan_area_and_windows (Mustek_Scanner * s)
{
  double pixels_per_mm = SANE_UNFIX (s->hw->dpi_range.max) / MM_PER_INCH;
  u_int8_t cmd[117], *cp;
  int i, dim;

  /* setup SCSI command (except length): */
  memset (cmd, 0, sizeof (cmd));
  cmd[0] = MUSTEK_SCSI_AREA_AND_WINDOWS;

  cp = cmd + 6;

  /* fill in frame header: */

  /* pixel unit and halftoning: */
  *cp++ = 0x8 | ((s->mode & MUSTEK_MODE_HALFTONE) ? 0x01 : 0x00);

  /* fill in scanning area: */
  STORE16 (cp, SANE_UNFIX (s->val[OPT_TL_X].w) * pixels_per_mm + 0.5);
  STORE16 (cp, SANE_UNFIX (s->val[OPT_TL_Y].w) * pixels_per_mm + 0.5);
  STORE16 (cp, SANE_UNFIX (s->val[OPT_BR_X].w) * pixels_per_mm + 0.5);
  STORE16 (cp, SANE_UNFIX (s->val[OPT_BR_Y].w) * pixels_per_mm + 0.5);

  dim = s->val[OPT_HALFTONE_DIMENSION].w;
  if (dim)
    {
      *cp++ = 0x40;		/* mark presence of user pattern */
      *cp++ = (dim << 4) | dim;	/* set pattern length */
      for (i = 0; i < dim * dim; ++i)
	*cp++ = s->val[OPT_HALFTONE_PATTERN].wa[i];
    }

  cmd[4] = (cp - cmd) - 6;
  return genscsi_cmd (s->fd, cmd, (cp - cmd), 0, 0);
}

static SANE_Status
mode_select (Mustek_Scanner * s, int color_code)
{
  int grain_code, speed_code;
  u_int8_t mode[19], *cp;

  /* the scanners use a funky code for the grain size, let's compute it: */
  grain_code = s->val[OPT_GRAIN_SIZE].w;
  if (grain_code > 7)
    grain_code = 7;		/* code 0 is 8x8, not 7x7 */
  grain_code = 7 - grain_code;

  /* same goes for speed: */
  for (speed_code = 0; speed_list[speed_code]; ++speed_code)
    {
      if (strcmp (speed_list[speed_code], s->val[OPT_SPEED].s) == 0)
	break;
    }
  if (speed_code > 4)
    speed_code = 4;
  else if (speed_code < 0)
    speed_code = 0;
  speed_code = 4 - speed_code;	/* 0 is fast, 4 is slow */

  memset (mode, 0, sizeof (mode));
  mode[0] = MUSTEK_SCSI_MODE_SELECT;

  /* set command length and resolution code: */
  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    {
      mode[4] = 0x0d;
      cp = mode + 17;
      STORE16 (cp, s->resolution_code);
    }
  else
    {
      mode[4] = 0x0b;
      mode[7] = s->resolution_code;
    }
  /* Use `pixel' unit, select whether user-defined half-tone pattern
     should be used and tell scanner what color this mode select
     command refers to (used for 1-pass scanners only; 3-pass scanners
     always use COLOR_CODE_GRAY).  */
  mode[6] = (0x8b | (s->val[OPT_HALFTONE_DIMENSION].w ? 0x10 : 0x00)
	     | (color_code << 5));
  mode[8] = encode_percentage (s, SANE_UNFIX (s->val[OPT_BRIGHTNESS].w), 3.0);
  mode[9] = encode_percentage (s, SANE_UNFIX (s->val[OPT_CONTRAST].w), 7.0);
  mode[10] = grain_code;
  mode[11] = speed_code;	/* lamp setting not supported yet */
  mode[12] = 0;			/* shadow param not used by Mustek */
  mode[13] = 0;			/* highlist param not used by Mustek */
  mode[14] = mode[15] = 0;	/* paperlength not used by Mustek */
  mode[16] = 0;			/* midtone param not used by Mustek */

  return genscsi_cmd (s->fd, mode, 6 + mode[4], 0, 0);
}

static SANE_Status
gamma_correction (Mustek_Scanner * s, int color_code)
{
  u_int8_t gamma[266];
  int i, table, len;

  table = s->pass;
  if ((s->mode & MUSTEK_MODE_COLOR)
      && (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS))
    table = color_code - 1;

  memset (gamma, 0, sizeof (gamma));
  gamma[0] = MUSTEK_SCSI_LOOKUP_TABLE;
  gamma[9] = (color_code << 6);

  /* According to Mustek, the only builtin gamma table is a linear
     table, so all we support here is user-defined gamma tables.  */
  len = s->gamma_table[table].length;
  if (len)
    {
      gamma[2] = 0x27;		/* indicate user-selected gamma table */
      gamma[7] = (len >> 8) & 0xff;	/* big endian! */
      gamma[8] = (len >> 0) & 0xff;
      for (i = 0; i < len; ++i)
	gamma[10 + i] = s->gamma_table[table].value[i];
      len += 2;
    }
  len += 10;
  return genscsi_cmd (s->fd, gamma, len, 0, 0);
}

static SANE_Status
start_scan (Mustek_Scanner * s)
{
  u_int8_t start[6];

  memset (start, 0, sizeof (start));
  start[0] = MUSTEK_SCSI_START_STOP;
  start[4] = 0x01;
  if (s->mode & MUSTEK_MODE_COLOR)
    {
      if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
	start[4] |= 0x20;
      else
	start[4] |= ((s->pass << 3) + 1);
    }
  /* or in single/multi bit: */
  start[4] |= (s->mode & MUSTEK_MODE_MULTIBIT) ? (1 << 6) : 0;

  if (!s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    /* or in expanded resolution bit: */
    start[4] |= (s->resolution_code & 0x100) >> 1;

  return genscsi_cmd (s->fd, start, sizeof (start), 0, 0);
}

static SANE_Status
stop_scan (Mustek_Scanner * s)
{
  return genscsi_cmd (s->fd, stop, sizeof (stop), 0, 0);
}

static SANE_Status
do_cancel (Mustek_Scanner * s)
{
  s->scanning = SANE_FALSE;
  s->pass = 0;

  if (s->pipe >= 0)
    {
      close (s->pipe);
      s->pipe = -1;
    }

  if (s->reader_pid > 0)
    {
      int exit_status;

      kill (s->reader_pid, SIGTERM);
      while (wait (&exit_status) != s->reader_pid);
      s->reader_pid = 0;
    }

  if (s->fd >= 0)
    {
      /* ensure child knows it's time to stop: */
      stop_scan (s);
      genscsi_close (s->fd);
      s->fd = -1;
    }

  return SANE_STATUS_CANCELLED;
}

/* Determine the CCD's distance between the primary color lines.  */
static SANE_Status
line_distance (Mustek_Scanner * s)
{
#ifdef LINE_DISTANCE_WORKAROUND
  int factor, res, color;
#endif
  SANE_Status status;
  u_int8_t result[5];
  int peak_res;
  size_t len;

  len = sizeof (result);
  status = genscsi_cmd (s->fd, distance, sizeof (distance), result, &len);
  if (status != SANE_STATUS_GOOD)
    return status;

  res = SANE_UNFIX (s->val[OPT_RESOLUTION].w) + 0.5;
  peak_res = SANE_UNFIX (s->hw->dpi_range.max) + 0.5;
#ifdef LINE_DISTANCE_WORKAROUND
  result[0] = 0xff;
  result[1] = 0xff;
  if (s->mode & MUSTEK_MODE_COLOR)
    {
      if (peak_res == 600)
	{
	  if (res < 72)
	    {
	      /* 1-71 dpi: */
	      result[0] = 4; result[1] = 0;
	      result[2] = 0; result[3] = 3; result[4] = 5;
	    }
	  else if (res <= 300)
	    {
	      /* 72-300 dpi: */
	      result[0] = 2; result[1] = 0;
	      result[2] = 0; result[3] = 5; result[4] = 9;
	    }
	  else
	    {
	      /*301-600 dpi: */
	      result[0] = 1; result[1] = 0;
	      result[2] = 0; result[3] = 9; result[4] = 17;
	    }
	}
      else if (peak_res == 800)
	{
	  if (res < 72)
	    {
	      /* 1-71 dpi: */
	      result[0] = 4; result[1] = 0;
	      result[2] = 0; result[3] = 3; result[4] = 5;
	    }
	  else if (res <= 400)
	    {
	      /* 72-400 dpi: */
	      result[0] = 2; result[1] = 0;
	      result[2] = 0; result[3] = 9; result[4] = 17;
	    }
	  else
	    {
	      /*401-800 dpi: */
	      result[0] = 1; result[1] = 0;
	      result[2] = 0; result[3] = 16; result[4] = 32;
	    }
	}
    }
#endif /* LINE_DISTANCE_WORKAROUND */

  factor = result[0] | (result[1] << 8);
  if (factor != 0xffff)
    {
      /* need to do line-distance adjustment ourselves... */

      s->ld.max_value = peak_res;

      if (factor == 0)
	{
	  if (res <= peak_res / 2)
	    res *= 2;
	}
      else
	res *= factor;
      s->ld.peak_res = res;
      for (color = 0; color < 3; ++color)
	{
	  s->ld.c[color].dist = result[2 + color];
	  s->ld.c[color].Qk = s->ld.max_value;
	}
    }
  return SANE_STATUS_GOOD;
}

static SANE_Status
get_image_status (Mustek_Scanner * s,
		  SANE_Int * bytes_per_line, SANE_Int * lines)
{
  u_int8_t result[6];
  SANE_Status status;
  size_t len;
  int busy;

  do
    {
      len = sizeof (result);
      status = genscsi_cmd (s->fd, get_status, sizeof (get_status),
			    result, &len);
      if (status != SANE_STATUS_GOOD)
	return status;

      busy = result[0];
      if (busy)
	usleep (100000);

      if (!s->scanning)
	return do_cancel (s);
    }
  while (busy);

  *bytes_per_line = result[1] | (result[2] << 8);
  *lines = result[3] | (result[4] << 8) | (result[5] << 16);

  return SANE_STATUS_GOOD;
}

static SANE_Status
backtrack_and_adf (Mustek_Scanner * s)
{
  u_int8_t backtrack[6];
  int code = 0x80;

  if (s->val[OPT_BACKTRACK].w)
    code |= 0x02;

  if (strcmp (s->val[OPT_SOURCE].s, "Automatic Document Feeder") == 0)
    code |= 0x01;
  else if (strcmp (s->val[OPT_SOURCE].s, "Transparency Adapter") == 0)
    code |= 0x04;

  memset (backtrack, 0, sizeof (backtrack));
  backtrack[0] = MUSTEK_SCSI_ADF_AND_BACKTRACK;
  backtrack[4] = code;

  return genscsi_cmd (s->fd, backtrack, sizeof (backtrack), 0, 0);
}

static void
fix_line_distance (Mustek_Scanner * s, int num_lines, int bpl,
		   u_int8_t * raw, u_int8_t * out)
{
  u_int8_t *raw_end = raw + num_lines * bpl;
  int bpc = bpl / 3;		/* bytes per color (per line) */
  int index[3];			/* index of the next output line for color C */
  int i, c;
  static const int color_seq[] =
  {1, 2, 0};			/* green, blue, red */

  /* Initialize the indices with the line distances that were returned
     by the CCD linedistance command.  We want to skip data for the
     first OFFSET rounds, so we initialize the indices to the negative
     of this offset.  */
  for (c = 0; c < 3; ++c)
    index[c] = -s->ld.c[c].dist;

  while (1)
    {
      for (i = 0; i < 3; ++i)
	{
	  c = color_seq[i];
	  if (index[c] < 0)
	    ++index[c];
	  else if (index[c] < num_lines)
	    {
	      s->ld.c[c].Qk += s->ld.peak_res;
	      if (s->ld.c[c].Qk > s->ld.max_value)
		{
		  s->ld.c[c].Qk -= s->ld.max_value;
		  memcpy (out + c * bpc + index[c] * bpl, raw, bpc);
		  ++index[c];
		  raw += bpc;
		  if (raw >= raw_end)
		    return;
		}
	    }
	}
    }
}

static SANE_Status
read_data (Mustek_Scanner * s, SANE_Byte * buf, int lines, int bpl)
{
  u_int8_t readlines[6];
  size_t len;

  memset (readlines, 0, sizeof (readlines));
  readlines[0] = MUSTEK_SCSI_READ_SCANNED_DATA;
  readlines[2] = (lines >> 16) & 0xff;
  readlines[3] = (lines >> 8) & 0xff;
  readlines[4] = (lines >> 0) & 0xff;
  len = lines * bpl;
  return genscsi_cmd (s->fd, readlines, sizeof (readlines), buf, &len);
}

static SANE_Status
init_options (Mustek_Scanner * s)
{
  int i;

  memset (s->opt, 0, sizeof (s->opt));
  memset (s->val, 0, sizeof (s->val));

  for (i = 0; i < NUM_OPTIONS; ++i)
    {
      s->opt[i].size = sizeof (SANE_Word);
      s->opt[i].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    }

  s->opt[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
  s->opt[OPT_NUM_OPTS].cap = SANE_CAP_SOFT_DETECT;
  s->val[OPT_NUM_OPTS].w = NUM_OPTIONS;

  /* "Mode" group: */

  s->opt[OPT_MODE_GROUP].title = "Scan Mode";
  s->opt[OPT_MODE_GROUP].desc = "";
  s->opt[OPT_MODE_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_MODE_GROUP].cap = 0;
  s->opt[OPT_MODE_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* scan mode */
  s->opt[OPT_MODE].name = SANE_NAME_SCAN_MODE;
  s->opt[OPT_MODE].title = SANE_TITLE_SCAN_MODE;
  s->opt[OPT_MODE].desc = SANE_DESC_SCAN_MODE;
  s->opt[OPT_MODE].type = SANE_TYPE_STRING;
  s->opt[OPT_MODE].size = 0;			/* unbounded */
  s->opt[OPT_MODE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
  s->opt[OPT_MODE].constraint.string_list = mode_list;
  s->val[OPT_MODE].s = strdup (mode_list[2]);

  /* resolution */
  s->opt[OPT_RESOLUTION].name = SANE_NAME_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].title = SANE_TITLE_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].desc = SANE_DESC_SCAN_RESOLUTION;
  s->opt[OPT_RESOLUTION].type = SANE_TYPE_FIXED;
  s->opt[OPT_RESOLUTION].unit = SANE_UNIT_DPI;
  s->opt[OPT_RESOLUTION].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_RESOLUTION].constraint.range = &s->hw->dpi_range;
  s->val[OPT_RESOLUTION].w = 18 << SANE_FIXED_SCALE_SHIFT;

  /* speed */
  s->opt[OPT_SPEED].name = SANE_NAME_SCAN_SPEED;
  s->opt[OPT_SPEED].title = SANE_TITLE_SCAN_SPEED;
  s->opt[OPT_SPEED].desc = SANE_DESC_SCAN_SPEED;
  s->opt[OPT_SPEED].type = SANE_TYPE_STRING;
  s->opt[OPT_SPEED].size = 0;			/* unbounded */
  s->opt[OPT_SPEED].constraint_type = SANE_CONSTRAINT_STRING_LIST;
  s->opt[OPT_SPEED].constraint.string_list = speed_list;
  s->val[OPT_SPEED].s = strdup (speed_list[2]);

  /* source */
  s->opt[OPT_SOURCE].name = SANE_NAME_SCAN_SOURCE;
  s->opt[OPT_SOURCE].title = SANE_TITLE_SCAN_SOURCE;
  s->opt[OPT_SOURCE].desc = SANE_DESC_SCAN_SOURCE;
  s->opt[OPT_SOURCE].type = SANE_TYPE_STRING;
  s->opt[OPT_SOURCE].size = 0;			/* unbounded */
  s->opt[OPT_SOURCE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
  s->opt[OPT_SOURCE].constraint.string_list = source_list;
  s->val[OPT_SOURCE].s = strdup (source_list[0]);

  /* backtrack */
  s->opt[OPT_BACKTRACK].name = SANE_NAME_BACKTRACK;
  s->opt[OPT_BACKTRACK].title = SANE_TITLE_BACKTRACK;
  s->opt[OPT_BACKTRACK].desc = SANE_DESC_BACKTRACK;
  s->opt[OPT_BACKTRACK].type = SANE_TYPE_BOOL;
  s->val[OPT_BACKTRACK].w = SANE_FALSE;

  /* "Geometry" group: */

  s->opt[OPT_GEOMETRY_GROUP].title = "Geometry";
  s->opt[OPT_GEOMETRY_GROUP].desc = "";
  s->opt[OPT_GEOMETRY_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_GEOMETRY_GROUP].cap = 0;
  s->opt[OPT_GEOMETRY_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* top-left x */
  s->opt[OPT_TL_X].name = SANE_NAME_SCAN_TL_X;
  s->opt[OPT_TL_X].title = SANE_TITLE_SCAN_TL_X;
  s->opt[OPT_TL_X].desc = SANE_DESC_SCAN_TL_X;
  s->opt[OPT_TL_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_X].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_X].constraint.range = &s->hw->x_range;
  s->val[OPT_TL_X].w = 0;

  /* top-left y */
  s->opt[OPT_TL_Y].name = SANE_NAME_SCAN_TL_Y;
  s->opt[OPT_TL_Y].title = SANE_TITLE_SCAN_TL_Y;
  s->opt[OPT_TL_Y].desc = SANE_DESC_SCAN_TL_Y;
  s->opt[OPT_TL_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_TL_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_TL_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_TL_Y].constraint.range = &s->hw->y_range;
  s->val[OPT_TL_Y].w = 0;

  /* bottom-right x */
  s->opt[OPT_BR_X].name = SANE_NAME_SCAN_BR_X;
  s->opt[OPT_BR_X].title = SANE_TITLE_SCAN_BR_X;
  s->opt[OPT_BR_X].desc = SANE_DESC_SCAN_BR_X;
  s->opt[OPT_BR_X].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_X].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_X].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_X].constraint.range = &s->hw->x_range;
  s->val[OPT_BR_X].w = s->hw->x_range.max;

  /* bottom-right y */
  s->opt[OPT_BR_Y].name = SANE_NAME_SCAN_BR_Y;
  s->opt[OPT_BR_Y].title = SANE_TITLE_SCAN_BR_Y;
  s->opt[OPT_BR_Y].desc = SANE_DESC_SCAN_BR_Y;
  s->opt[OPT_BR_Y].type = SANE_TYPE_FIXED;
  s->opt[OPT_BR_Y].unit = SANE_UNIT_MM;
  s->opt[OPT_BR_Y].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BR_Y].constraint.range = &s->hw->y_range;
  s->val[OPT_BR_Y].w = s->hw->y_range.max;

  /* "Enhancement" group: */

  s->opt[OPT_ENHANCEMENT_GROUP].title = "Enhancement";
  s->opt[OPT_ENHANCEMENT_GROUP].desc = "";
  s->opt[OPT_ENHANCEMENT_GROUP].type = SANE_TYPE_GROUP;
  s->opt[OPT_ENHANCEMENT_GROUP].cap = 0;
  s->opt[OPT_ENHANCEMENT_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

  /* grain-size */
  s->opt[OPT_GRAIN_SIZE].name = SANE_NAME_GRAIN_SIZE;
  s->opt[OPT_GRAIN_SIZE].title = SANE_TITLE_GRAIN_SIZE;
  s->opt[OPT_GRAIN_SIZE].desc = SANE_DESC_GRAIN_SIZE;
  s->opt[OPT_GRAIN_SIZE].type = SANE_TYPE_INT;
  s->opt[OPT_GRAIN_SIZE].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_GRAIN_SIZE].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_GRAIN_SIZE].constraint.word_list = grain_list;
  s->val[OPT_GRAIN_SIZE].w = grain_list[1];

  /* brightness */
  s->opt[OPT_BRIGHTNESS].name = SANE_NAME_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].title = SANE_TITLE_BRIGHTNESS;
  s->opt[OPT_BRIGHTNESS].desc = SANE_DESC_BRIGHTNESS
    "  This option is active for lineart/halftone modes only.  "
    "For multibit modes (grey/color) use the gamma-table(s).";
  s->opt[OPT_BRIGHTNESS].type = SANE_TYPE_FIXED;
  s->opt[OPT_BRIGHTNESS].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_BRIGHTNESS].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_BRIGHTNESS].constraint.range = &percentage_range;
  s->val[OPT_BRIGHTNESS].w = 0;

  /* contrast */
  s->opt[OPT_CONTRAST].name = SANE_NAME_CONTRAST;
  s->opt[OPT_CONTRAST].title = SANE_TITLE_CONTRAST;
  s->opt[OPT_CONTRAST].desc = SANE_DESC_CONTRAST
    "  This option is active for lineart/halftone modes only.  "
    "For multibit modes (grey/color) use the gamma-table(s).";
  s->opt[OPT_CONTRAST].type = SANE_TYPE_FIXED;
  s->opt[OPT_CONTRAST].unit = SANE_UNIT_PERCENT;
  s->opt[OPT_CONTRAST].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_CONTRAST].constraint.range = &percentage_range;
  s->val[OPT_CONTRAST].w = 0;

  /* grayscale gamma vector */
  s->opt[OPT_GAMMA_VECTOR].name = SANE_NAME_GAMMA_VECTOR;
  s->opt[OPT_GAMMA_VECTOR].title = SANE_TITLE_GAMMA_VECTOR;
  s->opt[OPT_GAMMA_VECTOR].desc = SANE_DESC_GAMMA_VECTOR;
  s->opt[OPT_GAMMA_VECTOR].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR].wa = &s->gamma_table[0].value[0];

  /* red gamma vector */
  s->opt[OPT_GAMMA_VECTOR_R].name = SANE_NAME_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].title = SANE_TITLE_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].desc = SANE_DESC_GAMMA_VECTOR_R;
  s->opt[OPT_GAMMA_VECTOR_R].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_R].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_R].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_R].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_R].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_R].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_R].wa = &s->gamma_table[0].value[0];

  /* green gamma vector */
  s->opt[OPT_GAMMA_VECTOR_G].name = SANE_NAME_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].title = SANE_TITLE_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].desc = SANE_DESC_GAMMA_VECTOR_G;
  s->opt[OPT_GAMMA_VECTOR_G].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_G].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_G].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_G].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_G].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_G].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_G].wa = &s->gamma_table[1].value[0];

  /* blue gamma vector */
  s->opt[OPT_GAMMA_VECTOR_B].name = SANE_NAME_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].title = SANE_TITLE_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].desc = SANE_DESC_GAMMA_VECTOR_B;
  s->opt[OPT_GAMMA_VECTOR_B].type = SANE_TYPE_INT;
  s->opt[OPT_GAMMA_VECTOR_B].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_GAMMA_VECTOR_B].unit = SANE_UNIT_NONE;
  s->opt[OPT_GAMMA_VECTOR_B].size = 256 * sizeof (SANE_Word);
  s->opt[OPT_GAMMA_VECTOR_B].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_GAMMA_VECTOR_B].constraint.range = &u8_range;
  s->val[OPT_GAMMA_VECTOR_B].wa = &s->gamma_table[2].value[0];

  /* halftone dimension */
  s->opt[OPT_HALFTONE_DIMENSION].name = SANE_NAME_HALFTONE_DIMENSION;
  s->opt[OPT_HALFTONE_DIMENSION].title = SANE_TITLE_HALFTONE_DIMENSION;
  s->opt[OPT_HALFTONE_DIMENSION].desc = SANE_DESC_HALFTONE_DIMENSION;
  s->opt[OPT_HALFTONE_DIMENSION].type = SANE_TYPE_INT;
  s->opt[OPT_HALFTONE_DIMENSION].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_HALFTONE_DIMENSION].unit = SANE_UNIT_PIXEL;
  s->opt[OPT_HALFTONE_DIMENSION].constraint_type = SANE_CONSTRAINT_WORD_LIST;
  s->opt[OPT_HALFTONE_DIMENSION].constraint.word_list = pattern_dimension_list;
  s->val[OPT_HALFTONE_DIMENSION].w = 0;

  /* halftone pattern */
  s->opt[OPT_HALFTONE_PATTERN].name = SANE_NAME_HALFTONE_PATTERN;
  s->opt[OPT_HALFTONE_PATTERN].title = SANE_TITLE_HALFTONE_PATTERN;
  s->opt[OPT_HALFTONE_PATTERN].desc = SANE_DESC_HALFTONE_PATTERN;
  s->opt[OPT_HALFTONE_PATTERN].type = SANE_TYPE_INT;
  s->opt[OPT_HALFTONE_PATTERN].size = 0;
  s->opt[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;
  s->opt[OPT_HALFTONE_PATTERN].constraint_type = SANE_CONSTRAINT_RANGE;
  s->opt[OPT_HALFTONE_PATTERN].constraint.range = &u8_range;
  s->val[OPT_HALFTONE_PATTERN].wa = s->halftone_pattern;

  return SANE_STATUS_GOOD;
}

/* This function is executed as a child process.  The reason this is
   executed as a subprocess is because some (most?) generic SCSI
   interfaces block a SCSI request until it has completed.  With a
   subprocess, we can let it block waiting for the request to finish
   while the main process can go about to do more important things
   (such as recognizing when the user presses a cancel button).

   WARNING: Since this is executed as a subprocess, it's NOT possible
   to update any of the variables in the main process (in particular
   the scanner state cannot be updated).  */
static int
reader_process (Mustek_Scanner * s, int fd)
{
  SANE_Byte * data, *extra = 0;
  int lines_per_buffer, bpl, x, y, bit;
  SANE_Status status;
  sigset_t sigterm_set;
  FILE *fp;

  sigemptyset (&sigterm_set);
  sigaddset (&sigterm_set, SIGTERM);

  fp = fdopen (fd, "w");
  if (!fp)
    return 1;

  bpl = s->params.bytes_per_line;

  if ((s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
      && ((s->mode & (MUSTEK_MODE_COLOR | MUSTEK_MODE_MULTIBIT))
	  == MUSTEK_MODE_COLOR))
    /* In single-bit, single-pass color mode we expand every bit of
       information into a byte (0x00 or 0xff).  */
    bpl /= 8;

  lines_per_buffer = GENSCSI_MAX_REQUEST_SIZE / bpl;
  if (!lines_per_buffer)
    return 2;			/* resolution is too high */

  data = malloc (lines_per_buffer * bpl);

  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    /* get temporary buffer for line-distance correction and/or bit
       expansion */
    extra = malloc (lines_per_buffer * bpl);


  for (s->line = 0; s->line < s->params.lines; s->line += lines_per_buffer)
    {
      if (s->line + lines_per_buffer > s->params.lines)
	/* do the last few lines: */
	lines_per_buffer = s->params.lines - s->line;

      sigprocmask (SIG_BLOCK, &sigterm_set, 0);
      status = read_data (s, data, lines_per_buffer, bpl);
      sigprocmask (SIG_UNBLOCK, &sigterm_set, 0);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG(1, "reader_process: read_data failed with status=%d\n", status);
	  return 3;
	}

      /* convert to pixel-interleaved format: */
      if (s->mode & MUSTEK_MODE_COLOR)
	{
	  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
	    {
	      SANE_Byte *src;

	      if (s->ld.max_value)
		/* need to correct for distance between r/g/b sensors: */
		fix_line_distance (s, lines_per_buffer, bpl, data, extra);
	      else
		memcpy (extra, data, lines_per_buffer * bpl);

	      src = extra;
	      if (s->mode & MUSTEK_MODE_MULTIBIT)
		{
		  /* each r/g/b sample is 8 bits in line-interleaved format */
		  for (y = 0; y < lines_per_buffer; ++y)
		    {
		      for (x = 0; x < bpl / 3; ++x)
			{
			  fputc (src[0 * bpl / 3 + x], fp);
			  fputc (src[1 * bpl / 3 + x], fp);
			  fputc (src[2 * bpl / 3 + x], fp);
			}
		      src += bpl;
		    }
		}
	      else
		{
		  /* each r/g/b/ sample is 1 bit in line-interleaved format */
#		  define EXPAND(b, v) (((v) & (1 << (bit))) ? 0xff : 0x00)
		  for (y = 0; y < lines_per_buffer; ++y)
		    {
		      for (x = 0; x < bpl / 3; ++x)
			{
			  for (bit = 7; bit >= 0; --bit)
			    {
			      fputc (EXPAND (bit, src[0 * bpl / 3 + x]), fp);
			      fputc (EXPAND (bit, src[1 * bpl / 3 + x]), fp);
			      fputc (EXPAND (bit, src[2 * bpl / 3 + x]), fp);
			    }
			}
		      src += bpl;
		    }
		}
	    }
	}
      else
	{
	  if (s->mode & MUSTEK_MODE_MULTIBIT)
	    fwrite (data, lines_per_buffer, bpl, fp);
	  else
	    {
	      /* in singlebit mode, the scanner returns 1 for black. ;-( */
	      int i;

	      for (i = 0; i < lines_per_buffer * bpl; ++i)
		fputc (~data[i], fp);
	    }
	}
    }
  fclose (fp);
  return 0;
}

SANE_Status
sane_init (SANE_Int * version_code)
{
  char dev_name[PATH_MAX];
  size_t len;
  FILE * fp;

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);

  fp = fopen (PATH_MUSTEK_CONFIG, "r");
  if (!fp)
    {
      /* default to /dev/scanner instead of insisting on config file */
      attach ("/dev/scanner", 0);
      return SANE_STATUS_GOOD;
    }

  while (fgets (dev_name, sizeof (dev_name), fp))
    {
      if (dev_name[0] == '#')		/* ignore line comments */
	continue;
      len = strlen (dev_name);
      if (dev_name[len - 1] == '\n')
	dev_name[--len] = '\0';

      if (!len)
	continue;			/* ignore empty lines */

      attach (dev_name, 0);
    }
  fclose (fp);
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_exit (void)
{
  Mustek_Device * dev, * next;

  for (dev = first_dev; dev; dev = next)
    {
      next = dev->next;
      free ((void *) dev->sane.name);
      free ((void *) dev->sane.model);
      free (dev);
    }
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_get_devices (const SANE_Device *** device_list)
{
  static const SANE_Device ** devlist = 0;
  Mustek_Device * dev;
  int i;

  if (devlist)
    free (devlist);

  devlist = malloc ((num_devices + 1) * sizeof (devlist[0]));
  if (!devlist)
    return SANE_STATUS_NO_MEM;

  i = 0;
  for (dev = first_dev; i < num_devices; dev = dev->next)
    devlist[i++] = &dev->sane;
  devlist[i++] = 0;

  *device_list = devlist;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const devicename, SANE_Handle * handle)
{
  Mustek_Device *dev;
  SANE_Status status;
  Mustek_Scanner *s;

  if (devicename[0])
    {
      for (dev = first_dev; dev; dev = dev->next)
	if (strcmp (dev->sane.name, devicename) == 0)
	  break;

      if (!dev)
	{
	  status = attach (devicename, &dev);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}
    }
  else
    /* empty devicname -> use first device */
    dev = first_dev;

  if (!dev)
    return SANE_STATUS_INVAL;

  s = malloc (sizeof (*s));
  if (!s)
    return SANE_STATUS_NO_MEM;
  memset (s, 0, sizeof (*s));
  s->fd = -1;
  s->pipe = -1;
  s->hw = dev;

  init_options (s);

  *handle = s;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_close (SANE_Handle handle)
{
  Mustek_Scanner *s = handle;

  if (s->scanning)
    sane_cancel (handle);

  free (handle);
  return SANE_STATUS_GOOD;
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  Mustek_Scanner *s = handle;

  if ((unsigned) option >= NUM_OPTIONS)
    return 0;
  return s->opt + option;
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *val, SANE_Int * info)
{
  Mustek_Scanner *s = handle;
  SANE_Status status;
  SANE_Word cap;

  if (info)
    *info = 0;

  if (s->scanning || option >= NUM_OPTIONS)
    return SANE_STATUS_INVAL;

  cap = s->opt[option].cap;

  if (!SANE_OPTION_IS_ACTIVE (cap))
    return SANE_STATUS_INVAL;

  if (action == SANE_ACTION_GET_VALUE)
    {
      switch (option)
	{
	  /* word options: */
	case OPT_RESOLUTION:
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	case OPT_NUM_OPTS:
	case OPT_GRAIN_SIZE:
	case OPT_BRIGHTNESS:
	case OPT_CONTRAST:
	case OPT_HALFTONE_DIMENSION:
	  *(SANE_Word *) val = s->val[option].w;
	  return SANE_STATUS_GOOD;

	  /* word-array options: */
	case OPT_HALFTONE_PATTERN:
	case OPT_GAMMA_VECTOR:
	case OPT_GAMMA_VECTOR_R:
	case OPT_GAMMA_VECTOR_G:
	case OPT_GAMMA_VECTOR_B:
	  memcpy (val, s->val[option].wa, s->opt[option].size);
	  return SANE_STATUS_GOOD;

	  /* string options: */
	case OPT_SPEED:
	case OPT_SOURCE:
	case OPT_MODE:
	  strcpy(val, s->val[option].s);
	  return SANE_STATUS_GOOD;
	}
    }
  else if (action == SANE_ACTION_SET_VALUE)
    {
      if (!SANE_OPTION_IS_SETTABLE (cap))
	return SANE_STATUS_INVAL;

      status = constrain_value (s, option, val, info);
      if (status != SANE_STATUS_GOOD)
	return status;

      switch (option)
	{
	  /* (mostly) side-effect-free word options: */
	case OPT_RESOLUTION:
	case OPT_TL_X:
	case OPT_TL_Y:
	case OPT_BR_X:
	case OPT_BR_Y:
	  if (info)
	    *info |= SANE_INFO_RELOAD_PARAMS;
	  /* fall through */
	case OPT_NUM_OPTS:
	case OPT_GRAIN_SIZE:
	case OPT_BRIGHTNESS:
	case OPT_CONTRAST:
	  s->val[option].w = *(SANE_Word *) val;
	  return SANE_STATUS_GOOD;

	  /* side-effect-free word-array options: */
	case OPT_HALFTONE_PATTERN:
	  memcpy (s->val[option].wa, val, s->opt[option].size);
	  return SANE_STATUS_GOOD;

	case OPT_GAMMA_VECTOR:
	  {
	    int i, num_tables = 1;

	    if (strcmp (s->val[OPT_MODE].s, "Color") == 0)
	      /* in color mode, use this vector as an "intensity" vector: */
	      num_tables = 3;

	    for (i = 0; i < num_tables; ++i)
	      {
		s->gamma_table[i].length = 256;
		memcpy (s->gamma_table[i].value, val, s->opt[option].size);
	      }
	  }

	case OPT_GAMMA_VECTOR_R:
	case OPT_GAMMA_VECTOR_G:
	case OPT_GAMMA_VECTOR_B:
	  s->gamma_table[option - OPT_GAMMA_VECTOR_R].length = 256;
	  memcpy (s->val[option].wa, val, s->opt[option].size);
	  return SANE_STATUS_GOOD;

	  /* side-effect-free single-string options: */
	case OPT_SPEED:
	case OPT_SOURCE:
	  if (s->val[option].s)
	    free (s->val[option].s);
	  s->val[option].s = strdup (val);
	  return SANE_STATUS_GOOD;

	  /* options with side-effects: */

	case OPT_MODE:
	  /* brightness/contrast are in binary modes only: */
	  {
	    int halftoning;

	    if (s->val[option].s)
	      free (s->val[option].s);
	    s->val[option].s = strdup (val);

	    if (info)
	      *info |= SANE_INFO_RELOAD_OPTIONS | SANE_INFO_RELOAD_PARAMS;

	    s->opt[OPT_BRIGHTNESS].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_CONTRAST].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_GAMMA_VECTOR].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_GAMMA_VECTOR_R].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_GAMMA_VECTOR_G].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_GAMMA_VECTOR_B].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_HALFTONE_DIMENSION].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;

	    halftoning = (strcmp (val, "Halftone") == 0
			  || strcmp (val, "Color Halftone") == 0);

	    if (halftoning
		|| strcmp (val, "Lineart") == 0
		|| strcmp (val, "Color Lineart") == 0)
	      {
		s->opt[OPT_BRIGHTNESS].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_CONTRAST].cap &= ~SANE_CAP_INACTIVE;

		if (halftoning)
		  {
		    s->opt[OPT_HALFTONE_DIMENSION].cap &= ~SANE_CAP_INACTIVE;
		    s->opt[OPT_HALFTONE_PATTERN].cap &= ~SANE_CAP_INACTIVE;
		  }
	      }

	    if (strcmp (val, "Gray") == 0)
	      s->opt[OPT_GAMMA_VECTOR].cap &= ~SANE_CAP_INACTIVE;
	    else if (strcmp (val, "Color") == 0)
	      {
		s->opt[OPT_GAMMA_VECTOR].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_GAMMA_VECTOR_R].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_GAMMA_VECTOR_G].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_GAMMA_VECTOR_B].cap &= ~SANE_CAP_INACTIVE;
	      }
	    return SANE_STATUS_GOOD;
	  }

	case OPT_HALFTONE_DIMENSION:
	  /* halftone pattern dimension affects halftone pattern option: */
	  {
	    unsigned dim = *(SANE_Word *) val;

	    s->val[option].w = dim;

	    if (info)
	      *info |= SANE_INFO_RELOAD_OPTIONS;

	    s->opt[OPT_HALFTONE_DIMENSION].cap |= SANE_CAP_INACTIVE;
	    s->opt[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;

	    if (dim > 0)
	      {
		s->opt[OPT_HALFTONE_DIMENSION].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_HALFTONE_PATTERN].cap &= ~SANE_CAP_INACTIVE;
		s->opt[OPT_HALFTONE_PATTERN].size = dim * sizeof (SANE_Word);
	      }
	    return SANE_STATUS_GOOD;
	  }
	}
    }
  return SANE_STATUS_INVAL;
}

SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  Mustek_Scanner *s = handle;

  if (!s->scanning)
    {
      double width, height, dpi;
      const char *mode;

      memset (&s->params, 0, sizeof (s->params));

      width = SANE_UNFIX (s->val[OPT_BR_X].w - s->val[OPT_TL_X].w);
      height = SANE_UNFIX (s->val[OPT_BR_Y].w - s->val[OPT_TL_Y].w);
      dpi = SANE_UNFIX (s->val[OPT_RESOLUTION].w);

      /* make best-effort guess at what parameters will look like once
         scanning starts.  */
      s->params.first_frame = (s->pass == 0);
      if (dpi > 0.0 && width > 0.0 && height > 0.0)
	{
	  double dots_per_mm = dpi / MM_PER_INCH;

	  s->params.pixels_per_line = width * dots_per_mm;
	  s->params.lines = height * dots_per_mm;
	}

      mode = s->val[OPT_MODE].s;
      if (strcmp (mode, "Lineart") == 0 || strcmp (mode, "Halftone") == 0)
	{
	  s->params.format = SANE_FRAME_GRAY;
	  s->params.bytes_per_line = (s->params.pixels_per_line + 7) / 8;
	  s->params.depth = 1;
	}
      else if (strcmp (mode, "Gray") == 0)
	{
	  s->params.format = SANE_FRAME_GRAY;
	  s->params.bytes_per_line = s->params.pixels_per_line;
	  s->params.depth = 8;
	}
      else
	{
	  /* all color modes are treated the same since lineart with 1
	     bit color pixels still results in 3 bytes/pixel.  */
	  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
	    {
	      s->params.format = SANE_FRAME_RGB;
	      s->params.bytes_per_line = 3 * s->params.pixels_per_line;
	      s->params.depth = 8;
	    }
	  else
	    {
	      switch (s->pass)
		{
		case 0:
		  s->params.format = SANE_FRAME_RED;
		  break;
		case 1:
		  s->params.format = SANE_FRAME_GREEN;
		  break;
		case 2:
		  s->params.format = SANE_FRAME_BLUE;
		  break;
		default:
		  break;
		}
	      s->params.bytes_per_line = s->params.pixels_per_line;
	      s->params.depth = 8;
	    }
	}
    }
  if (params)
    *params = s->params;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_start (SANE_Handle handle)
{
  Mustek_Scanner *s = handle;
  SANE_Status status;
  int fds[2];

  if (s->fd < 0)
    {
      const char *mode;

      /* First make sure we have a current parameter set.  Some of the
         parameters will be overwritten below, but that's OK.  */
      status = sane_get_parameters (s, 0);
      if (status != SANE_STATUS_GOOD)
	return status;

      /* translate options into s->mode for convenient access: */
      mode = s->val[OPT_MODE].s;
      if (strcmp (mode, "Lineart") == 0)
	s->mode = 0;
      else if (strcmp (mode, "Halftone") == 0)
	s->mode = MUSTEK_MODE_HALFTONE;
      else if (strcmp (mode, "Gray") == 0)
	s->mode = MUSTEK_MODE_MULTIBIT;
      else if (strcmp (mode, "Color Lineart") == 0)
	s->mode = MUSTEK_MODE_COLOR;
      else if (strcmp (mode, "Color Halftone") == 0)
	s->mode = MUSTEK_MODE_COLOR | MUSTEK_MODE_HALFTONE;
      else if (strcmp (mode, "Color") == 0)
	s->mode = MUSTEK_MODE_COLOR | MUSTEK_MODE_MULTIBIT;

      s->one_pass_color_scan = ((s->mode & MUSTEK_MODE_COLOR)
				&& (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS));
      s->resolution_code = encode_resolution (s);

      status = genscsi_open (s->hw->sane.name, &s->fd);
      if (status != SANE_STATUS_GOOD)
	return status;

      status = scan_area_and_windows (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = request_sense (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = backtrack_and_adf (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = request_sense (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;
    }

  if (s->one_pass_color_scan)
    {
      status = mode_select (s, MUSTEK_CODE_RED);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = mode_select (s, MUSTEK_CODE_GREEN);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = mode_select (s, MUSTEK_CODE_BLUE);
    }
  else
    status = mode_select (s, MUSTEK_CODE_GRAY);

  status = request_sense (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  s->scanning = SANE_TRUE;

  status = start_scan (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  status = request_sense (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  if (s->one_pass_color_scan)
    {
      status = gamma_correction (s, MUSTEK_CODE_RED);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      gamma_correction (s, MUSTEK_CODE_GREEN);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = gamma_correction (s, MUSTEK_CODE_BLUE);
    }
  else
    status = gamma_correction (s, MUSTEK_CODE_GRAY);

  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  status = request_sense (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  s->ld.max_value = 0;
  if (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
    {
      status = line_distance (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;

      status = request_sense (s);
      if (status != SANE_STATUS_GOOD)
	goto stop_scanner_and_return;
    }

  status = backtrack_and_adf (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  status = request_sense (s);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  status = get_image_status (s, &s->params.bytes_per_line, &s->params.lines);
  if (status != SANE_STATUS_GOOD)
    goto stop_scanner_and_return;

  if ((s->mode & (MUSTEK_MODE_COLOR | MUSTEK_MODE_MULTIBIT))
      == MUSTEK_MODE_COLOR
      && (s->hw->flags & MUSTEK_FLAG_SINGLE_PASS))
    /* In single-bit, single-pass color mode we expand every bit of
       information into a byte (0x00 or 0xff).  */
    s->params.bytes_per_line *= 8;

  s->params.pixels_per_line = s->params.bytes_per_line;
  if (s->mode & MUSTEK_MODE_COLOR)
    s->params.pixels_per_line /= 3;
  else if (!(s->mode & MUSTEK_MODE_MULTIBIT))
    s->params.pixels_per_line *= 8;

  s->line = 0;

  if (pipe (fds) < 0)
    return SANE_STATUS_IO_ERROR;

  s->reader_pid = fork ();
  if (s->reader_pid == 0)
    {
      sigset_t ignore_set;
      struct sigaction act;

      close (fds[0]);

      sigfillset (&ignore_set);
      sigdelset (&ignore_set, SIGTERM);
      sigprocmask (SIG_SETMASK, &ignore_set, 0);

      memset (&act, 0, sizeof (act));
      sigaction (SIGTERM, &act, 0);

      /* don't use exit() since that would run the atexit() handlers... */
      _exit (reader_process (s, fds[1]));
    }
  close (fds[1]);
  s->pipe = fds[0];

  return SANE_STATUS_GOOD;

stop_scanner_and_return:
  sane_cancel (s);
  return status;
}

SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data, SANE_Int max_length,
	   SANE_Int * length)
{
  Mustek_Scanner *s = handle;
  ssize_t nread;

  *length = 0;

  nread = read (s->pipe, data, max_length);

  if (!s->scanning)
    return do_cancel (s);

  if (nread < 0)
    {
      if (errno == EAGAIN)
	return SANE_STATUS_AGAIN;
      else
	{
	  do_cancel (s);
	  return SANE_STATUS_IO_ERROR;
	}
    }

  *length = nread;

  if (nread == 0 && ((s->hw->flags & MUSTEK_FLAG_SINGLE_PASS)
		     || !(s->mode & MUSTEK_MODE_COLOR)
		     || ++s->pass >= 3))
    {
      do_cancel (s);
      return SANE_STATUS_EOF;
    }
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_cancel (SANE_Handle handle)
{
  Mustek_Scanner *s = handle;

  if (s->reader_pid > 0)
    kill (s->reader_pid, SIGTERM);
  s->scanning = SANE_FALSE;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  Mustek_Scanner *s = handle;

  if (!s->scanning)
    return SANE_STATUS_INVAL;

  if (fcntl (s->pipe, F_SETFL, non_blocking ? O_NONBLOCK : 0) < 0)
    return SANE_STATUS_IO_ERROR;

  return SANE_STATUS_UNSUPPORTED;
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  Mustek_Scanner *s = handle;

  if (!s->scanning)
    return SANE_STATUS_INVAL;

  *fd = s->pipe;
  return SANE_STATUS_GOOD;
}
