/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a simple SANE backend that can be used to load
   images in PNM format.  At present, only raw portable-pixmaps can be
   read (raw ppm format).  */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <sane.h>
#include <sanei.h>
#include <saneopts.h>

#define MAGIC	(void *)0xab730324

#define NELEMS(a)	((int)(sizeof (a) / sizeof (a[0])))

static int is_open = 0;
static char filename[128] = "/tmp/input.ppm";
static SANE_Fixed bright = 0;
static SANE_Fixed contr = 0;
static SANE_Bool gray = SANE_FALSE;
static FILE *infile = NULL;

static const SANE_Range percentage_range =
{
  -100 << SANE_FIXED_SCALE_SHIFT,	/* minimum */
   100 << SANE_FIXED_SCALE_SHIFT,	/* maximum */
     0 << SANE_FIXED_SCALE_SHIFT	/* quantization */
};

static const SANE_Option_Descriptor sod[] =
{
  {
    SANE_NAME_NUM_OPTIONS,
    SANE_TITLE_NUM_OPTIONS,
    SANE_DESC_NUM_OPTIONS,
    SANE_TYPE_INT,
    SANE_UNIT_NONE,
    sizeof (SANE_Word),
    SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    {NULL}
  },
  {
    "",
    "File",
    "Selection of the file to load.",
    SANE_TYPE_GROUP,
    SANE_UNIT_NONE,
    0,
    0,
    SANE_CONSTRAINT_NONE,
    {NULL}
  },
  {
    SANE_NAME_FILE,
    SANE_TITLE_FILE,
    SANE_DESC_FILE,
    SANE_TYPE_STRING,
    SANE_UNIT_NONE,
    sizeof (filename),
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    {NULL}
  },
  {
    "",
    "Image Enhancement",
    "A few controls to enhance image while loading",
    SANE_TYPE_GROUP,
    SANE_UNIT_NONE,
    0,
    0,
    SANE_CONSTRAINT_NONE,
    {NULL}
  },
  {
    SANE_NAME_BRIGHTNESS,
    SANE_TITLE_BRIGHTNESS,
    SANE_DESC_BRIGHTNESS,
    SANE_TYPE_FIXED,
    SANE_UNIT_PERCENT,
    sizeof (SANE_Word),
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_RANGE,
    {(SANE_String_Const *) &percentage_range}	/* this is ANSI conformant! */
  },
  {
    SANE_NAME_CONTRAST,
    SANE_TITLE_CONTRAST,
    SANE_DESC_CONTRAST,
    SANE_TYPE_FIXED,
    SANE_UNIT_PERCENT,
    sizeof (SANE_Word),
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_RANGE,
    {(SANE_String_Const *) &percentage_range}	/* this is ANSI conformant! */
  },
  {
    "grayify",
    "Grayify",
    "Load the image as grayscale.",
    SANE_TYPE_BOOL,
    SANE_UNIT_NONE,
    sizeof (SANE_Word),
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    {NULL}
  },
  {
    "default-enhancements",
    "Defaults",
    "Load defaults for enhancement.",
    SANE_TYPE_BUTTON,
    SANE_UNIT_NONE,
    0,
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    {NULL}
  }
};

static SANE_Parameters parms =
{
  SANE_FRAME_RGB,
  1,				/* Always first frame, no multichannel */
  0,				/* Number of bytes returned per scan line: */
  0,				/* Number of pixels per scan line.  */
  0,				/* Number of lines for the current scan.  */
  8,				/* Number of bits per sample. */
};

/* This library is a demo implementation of a SANE backend.  It
   implements a virtual device, a PNM file-filter. */
SANE_Status
sane_init (SANE_Int * version_code)
{
  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_exit (void)
{
  return SANE_STATUS_GOOD;
}

/* Device select/open/close */

static const SANE_Device dev[] =
{
  {
    "0",
    "noname",
    "PNM file reader",
    "virtual device"
  },
  {
    "1",
    "noname",
    "PNM file reader",
    "virtual device"
  },
};

SANE_Status
sane_get_devices (const SANE_Device *** device_list)
{
  static const SANE_Device * devlist[] =
  {
    dev + 0, dev + 1, 0
  };

  *device_list = devlist;
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const devicename, SANE_Handle * handle)
{
  int i;

  if (!devicename[0])
    i = 0;
  else
    for (i = 0; i < NELEMS(dev); ++i)
      if (strcmp (devicename, dev[i].name) == 0)
	break;
  if (i >= NELEMS(dev))
    return SANE_STATUS_INVAL;

  if (is_open)
    return SANE_STATUS_DEVICE_BUSY;

  is_open = 1;
  *handle = MAGIC;
  return SANE_STATUS_GOOD;
}

extern SANE_Status
sane_close (SANE_Handle handle)
{
  if (handle != MAGIC)
    return SANE_STATUS_INVAL;	/* Unknown handle ... */
  is_open = 0;
  return SANE_STATUS_GOOD;
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  if (handle != MAGIC || !is_open)
    return NULL;		/* wrong device */
  if (option < 0 || option >= NELEMS(sod))
    return NULL;
  return &sod[option];
}

extern SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action, void *value,
		     SANE_Int * info)
{
  SANE_Int myinfo = 0;
  SANE_Status status;

  if (handle != MAGIC || !is_open)
    return SANE_STATUS_INVAL;	/* Unknown handle ... */

  if (option < 0 || option >= NELEMS(sod))
    return SANE_STATUS_INVAL;	/* Unknown option ... */

  switch (action)
    {
    case SANE_ACTION_SET_VALUE:
      status = sanei_constrain_value (sod + option, value, &myinfo);
      if (status != SANE_STATUS_GOOD)
	return status;

      switch (option)
	{
	case 2:
	  strcpy (filename, value);
	  if (access (filename, R_OK) == 0)
	    myinfo |= SANE_INFO_RELOAD_PARAMS;
	  break;
	case 4:
	  bright = *(SANE_Word *) value;
	  break;
	case 5:
	  contr = *(SANE_Word *) value;
	  break;
	case 6:
	  gray = !!*(SANE_Word *) value;
	  myinfo |= SANE_INFO_RELOAD_PARAMS;
	  break;
	case 7:
	  bright = contr = 0;
	  myinfo |= SANE_INFO_RELOAD_OPTIONS;
	  break;
	default:
	  return SANE_STATUS_INVAL;
	}
      break;

    case SANE_ACTION_GET_VALUE:
      switch (option)
	{
	case 0:
	  *(SANE_Word *) value = NELEMS(sod);
	  break;
	case 2:
	  strcpy (value, filename);
	  break;
	case 4:
	  *(SANE_Word *) value = bright;
	  break;
	case 5:
	  *(SANE_Word *) value = contr;
	  break;
	case 6:
	  *(SANE_Word *) value = gray;
	  break;
	case 7:
	  break;
	default:
	  return SANE_STATUS_INVAL;
	}
      break;

    case SANE_ACTION_SET_AUTO:
      return SANE_STATUS_UNSUPPORTED;	/* We are DUMB */
    }
  if (info)
    *info = myinfo;
  return SANE_STATUS_GOOD;
}

static void
get_line (char *buf, int len, FILE * f)
{
  do
    fgets (buf, len, f);
  while (*buf == '#');
}

static int
getparmfromfile (void)
{
  FILE *fn;
  int x, y;
  char buf[1024];

  parms.bytes_per_line = parms.pixels_per_line = parms.lines = 0;
  if ((fn = fopen (filename, "r")) == NULL)
    return -1;

  get_line (buf, sizeof (buf), fn);
  if (strncmp (buf, "P6", 2))
    {
      fclose (fn);
      return -1;
    }
  get_line (buf, sizeof (buf), fn);
  sscanf (buf, "%d %d", &x, &y);

  parms.bytes_per_line = x;
  parms.pixels_per_line = x;
  parms.lines = y;
  if (gray)
    parms.format = SANE_FRAME_GRAY;
  else
    {
      parms.format = SANE_FRAME_RGB;
      parms.bytes_per_line *= 3;
    }
  return 0;
}

extern SANE_Status
sane_get_parameters (SANE_Handle handle,
		     SANE_Parameters * params)
{
  int rc = SANE_STATUS_GOOD;
  if (handle != MAGIC || !is_open)
    rc = SANE_STATUS_INVAL;	/* Unknown handle ... */
  else if (getparmfromfile ())
    rc = SANE_STATUS_INVAL;
  *params = parms;
  return rc;
}

extern SANE_Status
sane_start (SANE_Handle handle)
{
  char buf[1024];

  if (handle != MAGIC || !is_open)
    return SANE_STATUS_INVAL;	/* Unknown handle ... */

  if (getparmfromfile ())
    return SANE_STATUS_INVAL;

  if (infile != NULL)
    {
      fclose (infile);
      infile = NULL;
      return SANE_STATUS_EOF;
    }
  if ((infile = fopen (filename, "r")) == NULL)
    return SANE_STATUS_INVAL;
  get_line (buf, sizeof (buf), infile);	/* Skip the header */
  get_line (buf, sizeof (buf), infile);
  /* Can there be comments after the last header ??? */
  get_line (buf, sizeof (buf), infile);

  return SANE_STATUS_GOOD;
}

extern SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data,
	   SANE_Int max_length, SANE_Int * length)
{
  int len, x, hlp;

  if (handle != MAGIC || !is_open)
    return SANE_STATUS_INVAL;	/* Unknown handle ... */

  if (feof (infile))
    return SANE_STATUS_EOF;

  if (!gray)
    len = fread (data, 1, max_length, infile);
  else
    {
      unsigned char buf[3];
      for (len = 0; len < max_length; len++)
	{
	  if (fread (buf, 1, sizeof (buf), infile) != sizeof (buf))
	    break;
	  *(data + len) = ((long) buf[0] + buf[1] + buf[2]) / 3;
	}
    }

  /* Do the transformations ... DEMO ONLY ! THIS MAKES NO SENSE ! */
  for (x = 0; x < len; x++)
    {
      hlp = *((unsigned char *) data + x) - 128;
      hlp *= (contr + (100 << SANE_FIXED_SCALE_SHIFT));
      hlp /= 100 << SANE_FIXED_SCALE_SHIFT;
      hlp += (bright >> SANE_FIXED_SCALE_SHIFT) + 128;
      if (hlp < 0)
	hlp = 0;
      if (hlp > 255)
	hlp = 255;
      *(data + x) = hlp;
    }

  *length = len;
  return SANE_STATUS_GOOD;
}

extern SANE_Status
sane_cancel (SANE_Handle handle)
{
  if (infile != NULL)
    {
      fclose (infile);
      infile = NULL;
    }
  return SANE_STATUS_GOOD;
}

extern SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  return SANE_STATUS_UNSUPPORTED;
}

extern SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  return SANE_STATUS_UNSUPPORTED;
}
